// slist.h

#ifndef LQ_SLIST_H
#define LQ_SLIST_H

#include "objlist.h"
#include "cdecls.h"

LQ_BEGIN_CDECLS;

/** \file slist.h
 * Usage Tip:
 * As an OOP implementation, the first parameter for any call will be 
 * the pointer to a list and be of type char*** where char** is the 
 * list object type itself.  The one exception to this rule is the 
 * slist_count(char** obj) call, which differs for compatiblity with 
 * args types of lists and for(...) loops while slist_getCount(char***) 
 * does the same thing but operates on the pointer to the object as do
 * the rest of the functions.

 * NOTE: While the type of object is a char** and can be read as any 
 * list of args can be (last item is a null pointer), it should be 
 * considered a read-only array and to delete it use slist_delete(&obj) 
 * because it's not really just a char** array and the strings are all 
 * created (strdup()-ed) internally.
  
 * See dirlist.h and dirlist.c for how to handle more complext arrays 
 * such as structs requiring destructors for fields within the object.  
 */

/// returns an initialized pointer to a list base
char** slist_new();

/// removes items and deletes the base
void slist_delete(char*** list);

/// returns string at index
char * slist_get(char*** plist, int index);

/// USE WITH CAUTION: sets string at index, does not free previous
void _slist_set(char*** plist, int index, void* item);

/// returns number of items in the list
int slist_count(char** list);

/// returns number of items in the list
int slist_getCount(char*** plist);

/// removes items, does not resize or delete the base
void slist_clear(char*** plist);

/// adds a new items at the end of the list
void slist_append(char*** plist, const char* s);

/// removes (and frees) item at index. Does not resize the list
void slist_remove(char*** plist, int index);

/// inserts new item at index
void slist_insert(char*** plist, const char* s, int index);

/// adds list list2 to list1, does not delete list that's added.
/// plist2 can be any standard null terminated arg list or slist.
void slist_addList(char*** plist1, char*** list2);

/// removes items from index to end of list
void slist_trim(char*** plist, int index);

/// removes first n items from list
void slist_shift(char*** plist, int n);

/// reads lines (or args list) from a file, returns an slist object
/// that must be deleted by the user.  (I.e., it creates the list)
int slist_fileRead(char*** plist, const char* filename);

/// writes lines (or args list) to a file, return 0 on success,
/// list doesn't have to be a literal slist object, any null 
/// terminated list not having newlines at end of lines will work.
int slist_fileWrite(char*** plist, const char* filename);

/// Loads list with strings from a piped command,
/// returns 0 on success
int slist_pipeRead(char*** plist, const char* pipecmd);

/// writes list through pipe, returns 0 on success.
int slist_pipeWrite(char*** plist, const char* syscmd);

LQ_END_CDECLS;

#endif // LQ_SLIST_H
 
