// slist.c -- lists of strings in the format of an args list, 0 terminated.

#include <malloc.h> // malloc and free
#include <string.h> // str*() funcs

#include "slist.h"
#include "cdecls.h"

LQ_BEGIN_CDECLS;

// returns an initialized pointer to a list base
char** slist_new() { return (char**)objlist_new();}

// returns string at index
char * slist_get(char*** plist, int index)
{
  if(index >= slist_getCount(plist))
    return 0;
  if(index < 0)
    return 0;
  return (*plist)[index];
}

// USE WITH CAUTION: sets string at index, does not free previous
void _slist_set(char*** plist, int index, void* item)
{
  if(index >= slist_getCount(plist))
    return;
  if(index < 0)
    return;
  (*plist)[index] = (char*)item;
}


// removes item and deletes the base
void slist_delete(char*** plist) {objlist_delete((OBJLIST*)plist);}

// returns number of items in the list
int slist_count(char** list) { return objlist_count((OBJLIST)list);}

// returns number of items in the list
int slist_getCount(char*** plist) { return objlist_getCount((OBJLIST*)plist);}

// removes items, does not resize or delete the base
void slist_clear(char*** plist) {objlist_clear((OBJLIST*)plist);}

// adds a new items at the end of the list
void slist_append(char*** plist, const char* s)
{
  const char* str = strdup(s);
  objlist_append((OBJLIST*)plist, (OBJLIST_ITEM*)str);
}

// removes (and frees) item at index. Does not resize the list
void slist_remove(char*** plist, int index) 
{
  objlist_remove((OBJLIST*)plist, index);
}


// adds list list2 to list1, does not delete list that's added.
// plist2 can be any standard null terminated arg list or slist.
void slist_addList(char*** plist1, char*** plist2)
{
  char** s = *plist2;
  while(*s)
    slist_append(plist1, *s++);
}

// inserts new item at index
void slist_insert(char*** plist, const char* s, int index)
{
  const char* str = strdup(s);
  objlist_insert((OBJLIST*)plist, (OBJLIST_ITEM*)str, index);
}

//removes items from index to end of list
void slist_trim(char*** plist, int index)
{
  objlist_trimRight((OBJLIST*)plist, index);
}

//removes first n items from list
void slist_shift(char*** plist, int n)
{
  int index = n - 1;
  objlist_trimLeft((OBJLIST*)plist, index);
}

// TODO: maybe...
// overwrites and appends from index to end of plist2, trimmed 
// to end of plist2 slist_moveList(plist1, index, plist2, delete_flg)

// remove one trailing newline char from a string
static void trim_newline(char* buf)
{
  char* p = buf + strlen(buf) -1;
  if(*p == '\n')
    *p = 0;
}

// Loads list with strings from a piped command,
// returns 0 on success
int slist_pipeRead(char*** plist, const char* pipecmd)
{
  size_t bufsize = 4096;
  char* buf = (char*) malloc(bufsize);
  char** list = *plist;
  FILE* fp = popen(pipecmd, "r");
  if(! fp) return 1;
  while(! feof(fp))
  {
    *buf = 0;
    getline(&buf, &bufsize, fp);
    
    // done if nothing is read
    if(! *buf)
      break;
    
    trim_newline(buf);
    slist_append(&list, buf);
  }
  fclose(fp);
  
  // save the result and return ok
  free(buf);
  *plist = list;
  return 0;
}

// writes list through pipe, returns 0 on success.
int slist_pipeWrite(char*** plist, const char* pipecmd)
{
  int i; // loop index
  char** list = *plist;
  FILE* fp = popen(pipecmd, "w");
  if(! fp) return 1;
  for(i = 0; i < slist_count(list); i++)
  {
    char* line = list[i]; // to see in debugger
    fprintf(fp, "%s\n", line);    
  }
  fclose(fp);
  return 0;
}

// reads lines (or args list) from a file, requires an initialized
// slist object.  Returns 0 on success.
int slist_fileRead(char*** plist, const char* filename)
{
  char** list = *plist;
  size_t n = 4096;
  char* buf = (char*)malloc(n);
  if(!buf) return 1;
  FILE* fp = fopen(filename, "r");
  if(fp)
  {
    while(!feof(fp))
    {
      *buf = 0;
      getline(&buf, &n, fp);
      if(!*buf)
        break;
      trim_newline(buf);
      slist_append(&list, buf);
    }
    fclose(fp);
  }
  free(buf);
  *plist = list;
  return 0;
}

// this keeps pipes from screwing up a debugger in the newer linuxes
static void _fputs(char* s, FILE* fp)
{
  fputs(s, fp);
}

// write lines (or args list) to a file, return 0 on success,
// list doesn't have to be a literal slist object, any null 
// terminated list not having newlines at end of lines will work.
int slist_fileWrite(char*** plist, const char* filename)
{
  int i = 0;
  char** list = *plist;
  FILE* fp = fopen(filename, "w");
  if(!fp) return 1; // can't write, 
  while(list[i])
    _fputs(list[i++], fp);
  fclose(fp);
  return 0;
}

LQ_END_CDECLS;
