// find-deps.cpp

// and some commonly used c headers
#include <stdio.h>    // printf, sprintf, etc.
#include <stdlib.h>   // exit()
#include <unistd.h>   // tons of stuff
#include <malloc.h>   // memory allocation
#include <string.h>   // strings and block comparisons and moves

// our own includes
#include "slist.h"
#include "reorder_args.h"
#include "version.h"
#include "filename.h"

void dbg(){}          // a breakpoint for kdbg.exec

//////////////////////////////////////////////////////
// prototypes section


// find and store names of elf executables in elflist,
// returns 0 on success
int find_elf(char*** pelflist, const char* prefix);

// find and store dependency libs in depslist, not including files
// in elflist, returns 0 on success
int find_deps(char*** depslist, const char* prefix, char** elflist);

// process recognized switches and return errcode if bad switch
// value.
int handle_switches(int n, char** argv);

// usage note
int usage(int errcode);

//////////////////////////////////////////////////////
// the program

// saved current directory path
char HERE[256];

// true if we want to know how many files were processed
int infoflg = 0;

// true if we only want to know which if any libs are unresolved.
int unresflg;

// number of pipe to ldd errors
int ldderrs = 0;

// name of currently running app
char* appname;

char* nopath(char* s);

int main(int argc, char** argv)
{
  dbg();
  
  appname = basename(argv[0]);
  
  int err = 0;
  char** elflist = slist_new();
  char** depslist = slist_new();  
  int offset = 0; // increases for each commandline switch
  
  // syntax checks  
  if((argc < 2) || (strcmp(argv[1], "--help") == 0))
    return usage(0);
      
  // switches
  offset = reorder_args(argc, argv);
  err = handle_switches(offset, argv);
  if(err)
    return err;
  
  // setup
  const char* prefix = argv[1 + offset];
  strcpy(HERE, getenv("PWD"));
  
  // get list of all elf executables files in prefix dir
  err = find_elf(&elflist, prefix);
  
  int num_elf = slist_count(elflist);

  if(num_elf == 0)
  {
    fprintf(stderr, "No elf binaries found.\n");
    return 1;
  }  
  
  // get list of deps in all elf files that are not in the
  // list already in the prefix dir
  err = find_deps(&depslist, prefix, elflist);
  
  int num_deps = slist_count(depslist);
  int missing_deps = 0;
  
  for(int i = 0; i < slist_count(depslist); i++)
  {
    // print and count unresolved and if not doing
    // unresolved only, printout resolved as well
    if(strstr(depslist[i], "UNRESOLVED:"))
    {
      missing_deps++;
      printf("%s\n", nopath(depslist[i]));
    }
    else if(!unresflg)
      printf("%s\n", nopath(depslist[i]));
  }
  
  if(infoflg)
  {
    const char* hline =
        "---------------------------------------------------";
    printf("%s\n", hline);
    printf("binaries found : %d\n"
        "ldd pipe errors: %d\n"
        "external deps  : %d\n"
        "unresolved deps: %d \n",
        num_elf, ldderrs, num_deps, missing_deps);
    printf("%s\n", hline);
  }
  
  slist_delete(&depslist);
  slist_delete(&elflist);
  return 0;
}

//////////////////////////////////////////////////////
// the library of functions and various utils

// utils - misc

// returns true if filename is an elf file
int is_elf(const char* fname)
{
  // Exclude dirs. Somehow the dir can open as a file and 
  // return the ELF header of a file it contains.
  
  if(filename_isdir(fname))
    return 0;
  
  FILE* fp = fopen(fname, "r");
  if(! fp) return 0; // error, file not found
  
  // read the first few bytes of the file and look for the
  // signature of a valid elf binary.
  char buf[8] = {0};
  fread(buf, 1, 4, fp);
  if(memcmp(buf, "\177ELF", 4) == 0)
    return 1;
  return 0;
}


/////////////////////////////////////////////////////////

const char* unres_str = "UNRESOLVED: ";

// find and store names of elf executables in elflist,
// returns 0 on success
int find_elf(char*** pelflist, const char* prefix)
{  
  // init the flag and the list which is an array of 
  // pointers terminated by a zero and save current
  // directory so we can always return to it.
  int err = 0;
  char pipcmd[512];
  char** old_elflist = *pelflist; /// remove after test
  char** new_elflist = slist_new();

  // Create a non-looping do-while block we can leave 
  // at the first error.  Functionally equivalent to a 
  // string of ORs using err as the flag.
  do
  {  
    char pipecmd[512];
    err = chdir(prefix);
    if(err) break;
  
    // get complete list of all executables, then
    // remove those not elf type.
    // sprintf(pipecmd, "find * -executable");
    // added all *.so files, some libs in rpm packages may not be executable -rs
    sprintf(pipecmd, "find * -executable -o -name \"*.so*\"");

    err = slist_pipeRead(&old_elflist, pipecmd);
    if(err)
    {
      slist_clear(&old_elflist);
      break;
    }

    // remove non-elf files
    for(int i = 0; i < slist_count(old_elflist); i++)
    {
      if(is_elf(old_elflist[i]))
        slist_append(&new_elflist, old_elflist[i]);
    }
  }while(0);
  if(err) return err;
  
  // replace old_elflist with new_elflist
  slist_delete(&old_elflist);
  *pelflist = new_elflist;
  
  // delete tmp;
  chdir(HERE);
  return 0;
}

void elfpaths_only(char*** plist)
{
  int state = 0;
  char** list = *plist;
  char** tmp = slist_new();
  
  for(int i = 0; i < slist_count(list); i++)
  {
    char buf[256];
    const char* pname;
//    sprintf(buf, list[i]);
    do
    {
      state = 0;
      const char* ip = list[i];
      char* op = buf;
      
      // skip whitespaces
      while(*ip <= ' ')
        ip++;
      
      // remove blank lines
      if(!*ip)
        break;
      
      // remove dynloading defaults
      if(*ip == '/')
        break;
      
      pname = ip;
      
      // remove lines not pointing to lib path
      ip = strstr(ip, " => ");
      if(!ip)
        break;
      
      ip += 4; // move to file path, should start with '/' or tab (if rpathed)
      
      // remove lines not having a lib path or "not found"
      if(*ip != '/')
      {
        if(strstr(ip, "not found"))
        {
          op += sprintf(op, unres_str);
          while(*pname > 32)
            *op++ = *pname++;
          *op = 0;
          state = true;
          break;
        }
        else
          break;
      }
      // copy to first non whitespace and terminate string
      while(*ip > ' ')
        *op++ = *ip++;
      *op = 0;
      
      state = 1; // got one
    }while(0);
      
    if(state)
      // tmp.add(buf);
      slist_append(&tmp, buf);
  }
  slist_delete(&list);
  *plist = tmp;
}

// find and store dependency libs in depslist, not including files
// in elflist, returns 0 on success
int find_deps(char*** pdepslist, const char* prefix, char** elflist)
{
  // init
  char** depslist = *pdepslist;
  int err = 0;
  char pipecmd[512];
  
  err = chdir(prefix);
  if(err) return 1;
  
  int len = slist_count(elflist);
  for(int i = 0; i < len; i++)
  {
    // get list of ldd libs loaded by each elf file and add
    // to the depslist, temporarily before removing dups.
    
    char** sublist = slist_new();
    
    sprintf(pipecmd, "/usr/bin/ldd %s", elflist[i]);
    err = slist_pipeRead(&sublist, pipecmd);
    
    // count ldd errors
    if(err)
      ldderrs++;
    
    elfpaths_only(&sublist);
    
    slist_addList(&depslist, &sublist);
    slist_delete(&sublist);
  }while(0);
  
  int ip;
  int op;
  
  // remove duplicates if any items to copy
  ip = 0;  
  for(ip = 0; ip < slist_count(depslist); ip++)
  {
    op = ip + 1;
    while(op < slist_count(depslist))
    {
      if(strcmp(depslist[op], depslist[ip]) == 0)
        slist_remove(&depslist, op);
      else
        op++;
    }
  }
  
  // remove self-resolved. The file in the list may 
  // be a path or unres_str
  for(int elf = 0; elf < slist_count(elflist); elf++)
  {
    char* elffile = elflist[elf];
    if(strstr(elffile, unres_str))
       elffile = elffile + strlen(unres_str);
    else 
      elffile = basename(elffile);
    for(int dep = 0; dep < slist_count(depslist); dep++)
    {
      char* depfile = depslist[dep];
      if(strstr(depfile, unres_str))
         depfile = depfile + strlen(unres_str);
      else
        depfile = basename(depslist[dep]);
      if(strcmp(elffile, depfile) == 0)
        slist_remove(&depslist, dep);
    }
  }

  *pdepslist = depslist;
  chdir(HERE);
  return 0;
}

int usage(int errcode)
{
#include "usage_str.dat"
             printf(usage_str); 
         return errcode;
}

// process recognized switches and return errcode if bad switch
// value.
int handle_switches(int n, char** argv)
{
  n++; // don't count argv[0] (app path)
  
  // we already know the first char is a '-' so we look for
  // substrings common to both long and short switches.
  
  for(int i = 1; i < n; i++)
  {
    // -v or --version
    if(strstr(argv[i], "-v"))
    {
      printf("%s version %s\n", appname, VERSION);
      exit(0);
    }
    
    // -i or --info
    else if(strstr(argv[i], "-i"))
      infoflg = 1;
    
    // find unresolved only
    else if (strstr(argv[i], "-u"))
      unresflg = 1;
    
    else
    {
      fprintf(stderr, "Unrecognized switch %s\n", argv[i]);
      return 1; // not a valid switch
    }
  }
  return 0; // ok
}

char* nopath(char* s)
{
  char* p = strrchr(s, '/');
  if(!p)
    return s;
  else 
    return p+1;
}

// if you must, here's a way to do this without linking explicitly. :-)
// #include "slist.cpp" // uncomment this line.
// COMPILE: g++ find-deps.cpp -o find-deps
